@@ -184,7 +184,8 @@ class SendReceiveListener(object):
         body += Template(signature).safe_substitute(vals)
         body += '\n'
         cmsg.set_payload(body, charset='utf-8')
-        bdata = cmsg.as_bytes(policy=emlpolicy)
+        cmsg.set_charset('utf-8')
+        bdata = cmsg.as_bytes(policy=email.policy.SMTP)
         destaddrs = [identity]
         alwaysbcc = self._config['main'].get('alwayscc')
         if alwaysbcc:
@@ -298,7 +299,7 @@ class SendReceiveListener(object):
         new_hdrval = re.sub(r'\n?\s+', ' ', decoded)
         return new_hdrval.strip()
 
-    def receive(self, jdata, resp) -> None:
+    def receive(self, jdata, resp, reflect: bool = False) -> None:
         servicename = self._config['main'].get('myname')
         if not servicename:
             servicename = 'Web Endpoint'
@@ -308,8 +309,8 @@ class SendReceiveListener(object):
             return
         logger.debug('Received a request for %s messages', len(umsgs))
 
-        diffre = re.compile(r'^(---.*\n\+\+\+|GIT binary patch|diff --git \w/\S+ \w/\S+)', flags=re.M | re.I)
-        diffstatre = re.compile(r'^\s*\d+ file.*\d+ (insertion|deletion)', flags=re.M | re.I)
+        diffre = re.compile(rb'^(---.*\n\+\+\+|GIT binary patch|diff --git \w/\S+ \w/\S+)', flags=re.M | re.I)
+        diffstatre = re.compile(rb'^\s*\d+ file.*\d+ (insertion|deletion)', flags=re.M | re.I)
 
         msgs = list()
         conn = self._engine.connect()
@@ -317,12 +318,13 @@ class SendReceiveListener(object):
         t_auth = sa.Table('auth', md, autoload=True, autoload_with=self._engine)
         mustdest = self._config['main'].get('mustdest')
         # First, validate all messages
-        seenid = identity = selector = None
+        seenid = identity = selector = validfrom = None
         for umsg in umsgs:
+            bdata = umsg.encode()
             try:
-                identity, selector, auth_id = self.validate_message(conn, t_auth, umsg.encode())
+                identity, selector, auth_id = self.validate_message(conn, t_auth, bdata)
             except patatt.NoKeyError as ex:  # noqa
-                self.send_error(resp, message='No matching record found, maybe you need to auth-verify first?')
+                self.send_error(resp, message='No matching key, please complete web auth first.')
                 return
             except Exception as ex:
                 self.send_error(resp, message='Signature validation failed: %s' % ex)
@@ -335,21 +337,22 @@ class SendReceiveListener(object):
                 self.send_error(resp, message='We only support a single signing identity across patch series.')
                 return
 
-            msg = email.message_from_string(umsg)
+            msg = email.message_from_bytes(bdata, policy=email.policy.SMTP)
             logger.debug('Checking sanity on message: %s', msg.get('Subject'))
             # Some quick sanity checking:
             # - Subject has to start with [PATCH
             # - Content-type may ONLY be text/plain
             # - Has to include a diff or a diffstat
             passes = True
-            if not msg.get('Subject', '').startswith('[PATCH '):
+            subject = self.clean_header(msg.get('Subject', ''))
+            if not subject.startswith('[PATCH'):
                 passes = False
             if passes:
                 cte = msg.get_content_type()
                 if cte.lower() != 'text/plain':
                     passes = False
             if passes:
-                payload = msg.get_payload()
+                payload = msg.get_payload(decode=True)
                 if not (diffre.search(payload) or diffstatre.search(payload)):
                     passes = False
 
@@ -362,6 +365,29 @@ class SendReceiveListener(object):
                 self.send_error(resp, message='Message is missing some required headers.')
                 return
 
+            # Make sure that From: matches the validated identity. We allow + expansion,
+            # such that foo+listname@example.com is allowed for foo@example.com
+            allfroms = utils.getaddresses([str(x) for x in msg.get_all('from')])
+            # Allow only a single From: address
+            if len(allfroms) > 1:
+                self.send_error(resp, message='Message may only contain a single From: address.')
+                return
+
+            fromaddr = allfroms[0][1]
+            if validfrom != fromaddr:
+                ldparts = fromaddr.split('@')
+                if len(ldparts) != 2:
+                    self.send_error(resp, message=f'Invalid address in From: {fromaddr}')
+                    return
+                lparts = ldparts[0].split('+', maxsplit=1)
+                toval = f'{lparts[0]}@{ldparts[1]}'
+                if toval != identity:
+                    self.send_error(resp, message=f'From header invalid for identity {identity}: {fromaddr}')
+                    return
+                # usually, all From: addresses will be the same, so use validfrom as a quick bypass
+                if validfrom is None:
+                    validfrom = fromaddr
+
             # Check that To/Cc have a mailing list we recognize
             alldests = utils.getaddresses([str(x) for x in msg.get_all('to', [])])
             alldests += utils.getaddresses([str(x) for x in msg.get_all('cc', [])])
@@ -393,13 +419,19 @@ class SendReceiveListener(object):
             bccaddrs.update([x[1] for x in utils.getaddresses([_bcc])])
 
         repo = listid = None
-        if 'public-inbox' in self._config and self._config['public-inbox'].get('repo'):
+        if 'public-inbox' in self._config and self._config['public-inbox'].get('repo') and not reflect:
             repo = self._config['public-inbox'].get('repo')
             listid = self._config['public-inbox'].get('listid')
             if not os.path.isdir(repo):
                 repo = None
 
-        logger.info('Sending %s messages for %s/%s', len(msgs), identity, selector)
+        if reflect:
+            logger.info('Reflecting %s messages back to %s', len(msgs), identity, selector)
+            sentaction = 'Reflected'
+        else:
+            logger.info('Sending %s messages for %s/%s', len(msgs), identity, selector)
+            sentaction = 'Sent'
+
         for msg, destaddrs in msgs:
             subject = self.clean_header(msg.get('Subject'))
             if repo:
@@ -411,7 +443,7 @@ class SendReceiveListener(object):
                 ezpi.add_rfc822(repo, pmsg)
                 logger.debug('Wrote %s to public-inbox at %s', subject, repo)
 
-            origfrom = self.clean_header(msg.get('From'))
+            origfrom = msg.get('From')
             origpair = utils.getaddresses([origfrom])[0]
             origaddr = origpair[1]
             # Does it match one of our domains
@@ -425,12 +457,17 @@ class SendReceiveListener(object):
                 fromaddr = origaddr
             else:
                 logger.debug('%s does not match mydomain, substitution required', origaddr)
-                fromaddr = frompair[1]
                 # We can't just send this as-is due to DMARC policies. Therefore, we set
                 # Reply-To and X-Original-From.
+                fromaddr = frompair[1]
                 origname = origpair[0]
                 if not origname:
                     origname = origpair[1]
+                delim = self._config['main'].get('from-recipient-delimiter', '+')
+                if delim and '@' in fromaddr:
+                    _flocal, _fdomain = fromaddr.split('@', maxsplit=1)
+                    _forig = origaddr.replace('@', '.')
+                    fromaddr = f'{_flocal}{delim}{_forig}@{_fdomain}'
                 msg.replace_header('From', f'{origname} via {servicename} <{fromaddr}>')
 
                 if msg.get('X-Original-From'):
@@ -442,21 +479,34 @@ class SendReceiveListener(object):
                 else:
                     msg.add_header('Reply-To', f'<{origpair[1]}>')
 
-                body = msg.get_payload()
-                # Parse it as a message and see if we get a From: header
-                cmsg = email.message_from_string(body)
-                if cmsg.get('From') is None:
-                    cmsg.add_header('From', origfrom)
-                    msg.set_payload(cmsg.as_string(policy=emlpolicy, maxheaderlen=0), charset='utf-8')
+                body = msg.get_payload(decode=True)
+                # Add a From: header (if there isn't already one), but only if it's a patch
+                if diffre.search(body):
+                    # Parse it as a message and see if we get a From: header
+                    cmsg = email.message_from_bytes(body, policy=emlpolicy)
+                    if cmsg.get('From') is None:
+                        newbody = 'From: ' + self.clean_header(origfrom) + '\n'
+                        if cmsg.get('Subject'):
+                            newbody += 'Subject: ' + self.clean_header(cmsg.get('Subject')) + '\n'
+                        if cmsg.get('Date'):
+                            newbody += 'Date: ' + self.clean_header(cmsg.get('Date')) + '\n'
+                        newbody += '\n' + body.decode()
+                        msg.set_payload(newbody, charset='utf-8')
+                        # If we have non-ascii content in the new body, force CTE to 8bit
+                        if msg['Content-Transfer-Encoding'] == '7bit' and not all(ord(char) < 128 for char in newbody):
+                            msg.set_charset('utf-8')
+                            msg.replace_header('Content-Transfer-Encoding', '8bit')
 
             if bccaddrs:
                 destaddrs.update(bccaddrs)
 
-            bdata = msg.as_string(policy=emlpolicy).encode()
-
             if not self._config['main'].getboolean('dryrun'):
-                smtp.sendmail(fromaddr, list(destaddrs), bdata)
-                logger.info('Sent: %s', subject)
+                bdata = msg.as_bytes(policy=email.policy.SMTP)
+                if reflect:
+                    smtp.sendmail(fromaddr, [identity], bdata)
+                else:
+                    smtp.sendmail(fromaddr, list(destaddrs), bdata)
+                logger.info('%s: %s', sentaction, subject)
             else:
                 logger.info('---DRYRUN MSG START---')
                 logger.info(msg)
@@ -467,8 +517,8 @@ class SendReceiveListener(object):
             # run it once after writing all messages
             logger.debug('Running public-inbox repo hook (if present)')
             ezpi.run_hook(repo)
-        logger.info('Sent %s messages for %s/%s', len(msgs), identity, selector)
-        self.send_success(resp, message=f'Sent {len(msgs)} messages for {identity}/{selector}')
+        logger.info('%s %s messages for %s/%s', sentaction, len(msgs), identity, selector)
+        self.send_success(resp, message=f'{sentaction} {len(msgs)} messages for {identity}/{selector}')
 
     def on_post(self, req, resp):
         if not req.content_length:
@@ -501,6 +551,9 @@ class SendReceiveListener(object):
         if action == 'receive':
             self.receive(jdata, resp)
             return
+        if action == 'reflect':
+            self.receive(jdata, resp, reflect=True)
+            return
 
         resp.status = falcon.HTTP_500
         resp.content_type = falcon.MEDIA_TEXT
